@@ -9,9 +9,9 @@
 #include "lzfu.h"
 #include "msg.h"
 
-#define OUTPUT_TEMPLATE "%s"
+#define OUTPUT_TEMPLATE "%s.%s"
 #define OUTPUT_KMAIL_DIR_TEMPLATE ".%s.directory"
-#define KMAIL_INDEX ".%s.index"
+#define KMAIL_INDEX "../.%s.index"
 #define SEP_MAIL_FILE_TEMPLATE "%i%s"
 
 // max size of the c_time char*. It will store the date of the email
@@ -33,7 +33,7 @@ void      write_email_body(FILE *f, char *body);
 void      removeCR(char *c);
 void      usage();
 void      version();
-char*     mk_kmail_dir(char* fname);
+void      mk_kmail_dir(char* fname);
 int       close_kmail_dir();
 void      mk_recurse_dir(char* dir);
 int       close_recurse_dir();
@@ -66,10 +66,10 @@ void      write_journal(FILE* f_output, pst_item* item);
 void      write_appointment(FILE* f_output, pst_item *item);
 void      create_enter_dir(struct file_ll* f, pst_item *item);
 void      close_enter_dir(struct file_ll *f);
+char*     quote_string(char *inp);
 
 const char*  prog_name;
 char*  output_dir = ".";
-char*  kmail_chdir = NULL;
 
 // Normal mode just creates mbox format files in the current directory. Each file is named
 // the same as the folder's name that it represents
@@ -761,18 +761,12 @@ void version() {
 }
 
 
-char *mk_kmail_dir(char *fname) {
-    //change to that directory
+void mk_kmail_dir(char *fname) {
     //make a directory based on OUTPUT_KMAIL_DIR_TEMPLATE
-    //allocate space for OUTPUT_TEMPLATE and form a char* with fname
-    //return that value
-    char *dir, *out_name, *index;
+    //change to that directory
+    char *dir, *index;
     int x;
     DEBUG_ENT("mk_kmail_dir");
-    if (kmail_chdir && chdir(kmail_chdir)) {
-        x = errno;
-        DIE(("mk_kmail_dir: Cannot change to directory %s: %s\n", kmail_chdir, strerror(x)));
-    }
     dir = pst_malloc(strlen(fname)+strlen(OUTPUT_KMAIL_DIR_TEMPLATE)+1);
     sprintf(dir, OUTPUT_KMAIL_DIR_TEMPLATE, fname);
     check_filename(dir);
@@ -782,8 +776,10 @@ char *mk_kmail_dir(char *fname) {
             DIE(("mk_kmail_dir: Cannot create directory %s: %s\n", dir, strerror(x)));
         }
     }
-    kmail_chdir = pst_realloc(kmail_chdir, strlen(dir)+1);
-    strcpy(kmail_chdir, dir);
+    if (chdir(dir)) {
+        x = errno;
+        DIE(("mk_kmail_dir: Cannot change to directory %s: %s\n", dir, strerror(x)));
+    }
     free (dir);
 
     //we should remove any existing indexes created by KMail, cause they might be different now
@@ -792,25 +788,16 @@ char *mk_kmail_dir(char *fname) {
     unlink(index);
     free(index);
 
-    out_name = pst_malloc(strlen(fname)+strlen(OUTPUT_TEMPLATE)+1);
-    sprintf(out_name, OUTPUT_TEMPLATE, fname);
     DEBUG_RET();
-    return out_name;
 }
 
 
 int close_kmail_dir() {
-    // change ..
     int x;
     DEBUG_ENT("close_kmail_dir");
-    if (kmail_chdir) { //only free kmail_chdir if not NULL. do not change directory
-        free(kmail_chdir);
-        kmail_chdir = NULL;
-    } else {
-        if (chdir("..")) {
-            x = errno;
-            DIE(("close_kmail_dir: Cannot move up dir (..): %s\n", strerror(x)));
-        }
+    if (chdir("..")) {
+        x = errno;
+        DIE(("close_kmail_dir: Cannot move up dir (..): %s\n", strerror(x)));
     }
     DEBUG_RET();
     return 0;
@@ -874,7 +861,7 @@ void mk_recurse_dir(char *dir) {
             DIE(("mk_recurse_dir: Cannot create directory %s: %s\n", dir, strerror(x)));
         }
     }
-    if (chdir (dir)) {
+    if (chdir(dir)) {
         x = errno;
         DIE(("mk_recurse_dir: Cannot change to directory %s: %s\n", dir, strerror(x)));
     }
@@ -965,7 +952,7 @@ int close_separate_dir() {
 
 void mk_separate_file(struct file_ll *f, int32_t t, char *extension, int openit) {
     DEBUG_ENT("mk_separate_file");
-    DEBUG_INFO(("opening next file to save email\n"));
+    DEBUG_INFO(("opening next file to save email type %s\n", item_type_to_name(t)));
     if (f->item_count > 999999999) { // bigger than nine 9's
         DIE(("mk_separate_file: The number of emails in this folder has become too high to handle\n"));
     }
@@ -1165,6 +1152,32 @@ void write_embedded_message(FILE* f_output, pst_item_attach* attach, char *bound
     DEBUG_RET();
 }
 
+/**
+ * Backslash-escape quotes and backslashes in the given string.
+ */
+char *quote_string(char *inp) {
+    int i = 0;
+    int count = 0;
+    char *curr = inp;
+    while (*curr) {
+        *curr++;
+        if (*curr == '\"' || *curr == '\\') {
+            count++;
+        }
+        i++;
+    }
+    char *res = malloc(i + count + 1);
+    char *curr_in = inp;
+    char *curr_out = res;
+    while (*curr_in) {
+        if (*curr_in == '\"' || *curr_in == '\\') {
+            *curr_out++ = '\\';
+        }
+        *curr_out++ = *curr_in++;
+    }
+    *curr_out = '\0';
+    return res;
+}
 
 void write_inline_attachment(FILE* f_output, pst_item_attach* attach, char *boundary, pst_file* pst)
 {
@@ -1196,8 +1209,14 @@ void write_inline_attachment(FILE* f_output, pst_item_attach* attach, char *boun
     if (attach->filename2.str) {
         // use the long filename, converted to proper encoding if needed.
         // it is already utf8
+        char *escaped = quote_string(attach->filename2.str);
         pst_rfc2231(&attach->filename2);
-        fprintf(f_output, "Content-Disposition: attachment; \n        filename*=%s\n\n", attach->filename2.str);
+        fprintf(f_output, "Content-Disposition: attachment; \n        filename*=%s;\n", attach->filename2.str);
+        // Also include the (escaped) utf8 filename in the 'filename' header directly - this is not strictly valid
+        // (since this header should be ASCII) but is almost always handled correctly (and in fact this is the only
+        // way to get MS Outlook to correctly read a UTF8 filename, AFAICT, which is why we're doing it).
+        fprintf(f_output, "        filename=\"%s\"\n\n", escaped);
+        free(escaped);
     }
     else if (attach->filename1.str) {
         // short filename never needs encoding
@@ -1214,6 +1233,17 @@ void write_inline_attachment(FILE* f_output, pst_item_attach* attach, char *boun
 }
 
 
+int  header_match(char *header, char*field) {
+    int n = strlen(field);
+    if (strncasecmp(header, field, n) == 0) return 1;   // tag:{space}
+    if ((field[n-1] == ' ') && (strncasecmp(header, field, n-1) == 0)) {
+        char *crlftab = "\r\n\t";
+        DEBUG_INFO(("Possible wrapped header = %s\n", header));
+        if (strncasecmp(header+n-1, crlftab, 3) == 0) return 1; // tag:{cr}{lf}{tab}
+    }
+    return 0;
+}
+
 int  valid_headers(char *header)
 {
     // headers are sometimes really bogus - they seem to be fragments of the
@@ -1222,26 +1252,22 @@ int  valid_headers(char *header)
     // there are surely others. the problem is - given an arbitrary character
     // string, is it a valid (or even reasonable) set of rfc822 headers?
     if (header) {
-        if ((strncasecmp(header, "Content-Type: ",                  14) == 0) ||
-            (strncasecmp(header, "Date: ",                           6) == 0) ||
-            (strncasecmp(header, "From: ",                           6) == 0) ||
-            (strncasecmp(header, "MIME-Version: ",                  14) == 0) ||
-            (strncasecmp(header, "Microsoft Mail Internet Headers", 31) == 0) ||
-            (strncasecmp(header, "Received: ",                      10) == 0) ||
-            (strncasecmp(header, "Return-Path: ",                   13) == 0) ||
-            (strncasecmp(header, "Subject: ",                        9) == 0) ||
-            (strncasecmp(header, "To: ",                             4) == 0) ||
-            (strncasecmp(header, "X-ASG-Debug-ID: ",                16) == 0) ||
-            (strncasecmp(header, "X-Barracuda-URL: ",               17) == 0) ||
-            (strncasecmp(header, "X-x: ",                            5) == 0)) {
-            return 1;                                 
-        }
-        else {
-            if (strlen(header) > 2) {
-                DEBUG_INFO(("Ignore bogus headers = %s\n", header));
-            }
-            return 0;
+        if (header_match(header, "Content-Type: "                 )) return 1;
+        if (header_match(header, "Date: "                         )) return 1;
+        if (header_match(header, "From: "                         )) return 1;
+        if (header_match(header, "MIME-Version: "                 )) return 1;
+        if (header_match(header, "Microsoft Mail Internet Headers")) return 1;
+        if (header_match(header, "Received: "                     )) return 1;
+        if (header_match(header, "Return-Path: "                  )) return 1;
+        if (header_match(header, "Subject: "                      )) return 1;
+        if (header_match(header, "To: "                           )) return 1;
+        if (header_match(header, "X-ASG-Debug-ID: "               )) return 1;
+        if (header_match(header, "X-Barracuda-URL: "              )) return 1;
+        if (header_match(header, "X-x: "                          )) return 1;
+        if (strlen(header) > 2) {
+            DEBUG_INFO(("Ignore bogus headers = %s\n", header));
         }
+        return 0;
     }
     else return 0;
 }
@@ -2181,9 +2207,16 @@ void create_enter_dir(struct file_ll* f, pst_item *item)
     strcpy(f->dname, item->file_as.str);
 
     DEBUG_ENT("create_enter_dir");
-    if (mode == MODE_KMAIL)
-        f->name[0] = mk_kmail_dir(item->file_as.str);
-    else if (mode == MODE_RECURSE) {
+    if (mode == MODE_KMAIL) {
+        int32_t t;
+        mk_kmail_dir(item->file_as.str);
+        for (t=0; t<PST_TYPE_MAX; t++) {
+            if (t == reduced_item_type(t)) {
+                f->name[t] = (char*) pst_malloc(strlen(item->file_as.str)+strlen(OUTPUT_TEMPLATE)+30);
+                sprintf(f->name[t], OUTPUT_TEMPLATE, item->file_as.str, item_type_to_name(t));
+            }
+        }
+    } else if (mode == MODE_RECURSE) {
         int32_t t;
         mk_recurse_dir(item->file_as.str);
         for (t=0; t<PST_TYPE_MAX; t++) {
@@ -2198,12 +2231,23 @@ void create_enter_dir(struct file_ll* f, pst_item *item)
         }
     } else if (mode == MODE_SEPARATE) {
         // do similar stuff to recurse here.
+        int32_t t;
         mk_separate_dir(item->file_as.str);
-        f->name[0] = (char*) pst_malloc(file_name_len);
-        memset(f->name[0], 0, file_name_len);
+        for (t=0; t<PST_TYPE_MAX; t++) {
+            if (t == reduced_item_type(t)) {
+                f->name[t] = (char*) pst_malloc(file_name_len);
+                memset(f->name[t], 0, file_name_len);
+            }
+        }
     } else {
-        f->name[0] = (char*) pst_malloc(strlen(item->file_as.str)+strlen(OUTPUT_TEMPLATE)+1);
-        sprintf(f->name[0], OUTPUT_TEMPLATE, item->file_as.str);
+        // MODE_NORMAL
+        int32_t t;
+        for (t=0; t<PST_TYPE_MAX; t++) {
+            if (t == reduced_item_type(t)) {
+                f->name[t] = (char*) pst_malloc(strlen(item->file_as.str)+strlen(OUTPUT_TEMPLATE)+30);
+                sprintf(f->name[t], OUTPUT_TEMPLATE, item->file_as.str, item_type_to_name(t));
+            }
+        }
     }
 
     if (mode != MODE_SEPARATE) {
@@ -2218,12 +2262,11 @@ void create_enter_dir(struct file_ll* f, pst_item *item)
                     check_filename(temp);
                     while ((f->output[t] = fopen(temp, "r"))) {
                         DEBUG_INFO(("need to increase filename because one already exists with that name\n"));
-                        DEBUG_INFO(("- increasing it to %s%d\n", f->name, x));
                         x++;
-                        sprintf(temp, "%s%08d", f->name, x);
-                        DEBUG_INFO(("- trying \"%s\"\n", f->name));
+                        sprintf(temp, "%s%08d", f->name[t], x);
+                        DEBUG_INFO(("- bump file name and try \"%s\"\n", temp));
                         if (x == 99999999) {
-                            DIE(("create_enter_dir: Why can I not create a folder %s? I have tried %i extensions...\n", f->name, x));
+                            DIE(("create_enter_dir: Why can I not create a folder %s? I have tried %i extensions...\n", f->name[t], x));
                         }
                         fclose(f->output[t]);
                     }
@@ -2260,17 +2303,19 @@ void close_enter_dir(struct file_ll *f)
     for (t=0; t<PST_TYPE_MAX; t++) {
         if (f->output[t]) {
             if (mode == MODE_SEPARATE) DEBUG_WARN(("close_enter_dir finds open separate file\n"));
-            struct stat st;
             fclose(f->output[t]);
+            f->output[t] = NULL;
+        }
+        if (f->name[t]) {
+            struct stat st;
             stat(f->name[t], &st);
             if (!st.st_size) {
                 DEBUG_WARN(("removing empty output file %s\n", f->name[t]));
                 remove(f->name[t]);
             }
-            f->output[t] = NULL;
+            free(f->name[t]);
+            f->name[t] = NULL;
         }
-        free(f->name[t]);
-        f->name[t] = NULL;
     }
     free(f->dname);
 
